"""workorder hooks: compute progress"""

__docformat__ = "restructuredtext en"

from cubicweb.predicates import is_instance
from cubicweb.server import hook

WORKORDER_FINAL_STATES = set((u'done',))


class UpdateProgressOp(hook.DataOperationMixIn, hook.Operation):
    def precommit_event(self):
        for eid, compute in self.get_data():
            if not self.cnx.deleted_in_transaction(eid):
                entity = self.cnx.entity_from_eid(eid)
                entity.update_progress(compute)


class UpdateProgressOnWorkOrderStatusChange(hook.Hook):
    """when a ticket status change and it's identical to another one, change the
    state of the other one as well
    """
    __regid__ = 'workorder.progress.status_change'
    __select__ = hook.Hook.__select__ & is_instance('TrInfo')
    events = ('after_add_entity',)

    def __call__(self):
        trinfo = self.entity
        forentity = trinfo.for_entity
        if forentity.e_schema == 'WorkOrder' \
                and trinfo.new_state.name in WORKORDER_FINAL_STATES:
            forentity.set_attributes(progress_todo=0)


class UpdateProgressOnWorkOrderModification(hook.Hook):
    __regid__ = 'workorder.progress.workorder'
    __select__ = hook.Hook.__select__ & is_instance('WorkOrder')
    events = ('before_add_entity', 'before_update_entity', )

    def __call__(self):
        edited = self.entity.cw_edited
        if 'budget' in edited:
            if self.entity.progress_done is None:
                edited['progress_done'] = 0
            if self.entity.progress_todo is None:
                edited['progress_todo'] = edited['budget']
            UpdateProgressOp.get_instance(self._cw).add_data(
                (self.entity.eid, False))
        elif 'progress_todo' in edited or 'progress_done' in edited:
            UpdateProgressOp.get_instance(self._cw).add_data(
                (self.entity.eid, False))


class UpdateProgressOnSplitIntoChange(hook.Hook):
    __regid__ = 'workorder.progress.split_into'
    __select__ = (hook.Hook.__select__
                  & hook.match_rtype('split_into', frometypes=('Order',)))
    events = ('after_add_relation', 'after_delete_relation',)

    def __call__(self):
        UpdateProgressOp.get_instance(self._cw).add_data((self.eidfrom, True))



# pylint: disable-msg=W0622
"""cubicweb-workorder application packaging information"""

modname = 'workorder'
distname = 'cubicweb-%s' % modname

numversion = (0, 12, 1)
version = '.'.join(str(num) for num in numversion)

license = 'LGPL'
description = 'workorder component for the CubicWeb framework'
web = 'http://www.cubicweb.org/project/%s' % distname
author = 'Logilab'
author_email = 'contact@logilab.fr'
classifiers = [
           'Environment :: Web Environment',
           'Framework :: CubicWeb',
           'Programming Language :: Python',
           'Programming Language :: JavaScript',
]

__depends__ = {'cubicweb': '>= 3.24.0',
               'cubicweb-iprogress': '>= 0.1.1',
               }



"""template automatic tests"""

import unittest

from datetime import date, timedelta

from cubicweb.devtools.testlib import AutomaticWebTest, CubicWebTC


class AutomaticWebTest(AutomaticWebTest):

    def to_test_etypes(self):
        return set(('Order', 'WorkOrder'))

    def list_startup_views(self):
        return ()


class BasicEntitiesTC(CubicWebTC):

    def setup_database(self):
        """
        create 1 Order and 3 WorkOrder
        """
        with self.admin_access.repo_cnx() as cnx:
            # WorkOrder
            self.woben = cnx.create_entity('WorkOrder', title=u'Ben',
                                           begin_date=date(2001, 9, 11),
                                           end_date=date(2011, 5, 2)).eid
            self.woempty = cnx.create_entity('WorkOrder', title=u'Empty').eid
            self.woforrest = cnx.create_entity('WorkOrder', title=u'Forrest',
                                               begin_date=date.today() - timedelta(1),
                                               end_date=date.today() + timedelta(15)).eid
            # Order
            self.order = cnx.create_entity('Order', title=u'Coltrane', budget=100.0,
                                           split_into=(self.woben, self.woforrest,
                                                       self.woempty)).eid
            cnx.commit()

    def test_milestone_adapter(self):
        with self.admin_access.repo_cnx() as cnx:
            # Test dates of an empty workorder.
            empty = cnx.entity_from_eid(self.woempty)
            ms_empty = empty.cw_adapt_to('IMileStone')
            self.assertEqual(ms_empty.initial_prevision_date(),
                             date.today())

            # Test dates of an passed workorder, i.e. end date < today.
            ben = cnx.entity_from_eid(self.woben)
            ms_ben = ben.cw_adapt_to('IMileStone')
            self.assertEqual(ms_ben.initial_prevision_date(),
                             ben.end_date)
            self.assertEqual(ms_ben.eta_date(), date.today())

            # Test dates of an 'in progress' (not the state!) workorder, i.e. end
            # date > today
            forrest = cnx.entity_from_eid(self.woforrest)
            ms_forrest = forrest.cw_adapt_to('IMileStone')
            self.assertEqual(ms_forrest.initial_prevision_date(),
                             forrest.end_date)
            self.assertEqual(ms_forrest.eta_date(),
                             forrest.end_date)


if __name__ == '__main__':
    unittest.main()



from datetime import date

from cubicweb.entities import AnyEntity, fetch_config
from cubicweb.predicates import is_instance

from cubicweb_iprogress.entities import IMileStoneAdapter, IProgressAdapter


class Order(AnyEntity):
    __regid__ = 'Order'
    fetch_attrs, cw_fetch_order = fetch_config(('title', 'date'))

    def update_progress(self, compute=True):
        """callable only on the server side"""
        if compute:
            workorders = self.split_into
            self.cw_set(
                budget=sum(wod.budget or 0 for wod in workorders),
                progress_done=sum(wod.progress_done or 0 for wod in workorders),
                progress_todo=sum(wod.progress_todo or 0 for wod in workorders))


class WorkOrder(AnyEntity):
    __regid__ = 'WorkOrder'

    fetch_attrs, cw_fetch_order = fetch_config(('title', 'description_format',
                                                'description', 'budget',
                                                'begin_date', 'end_date', 'in_state'))

    def dc_title(self):
        return self.title

    def dc_long_title(self):
        return u'%s - %s' % (self.order.dc_title(), self.title)

    @property
    def order(self):
        return self.reverse_split_into[0]

    def update_progress(self, compute=True):
        """callable only on the server side"""
        for order in self.reverse_split_into:
            order.update_progress(True)


class WorkOrderIMileStoneAdapter(IMileStoneAdapter):
    __select__ = is_instance('WorkOrder')
    parent_type = 'Order'

    def get_main_task(self):
        return self.entity.order

    def initial_prevision_date(self):
        """returns the initial expected end of the milestone

        Warning: not taken into accound WorkOrder transition/state.
        """
        return self.entity.end_date or date.today()

    def eta_date(self):
        """returns expected date of completion based on what remains
        to be done

        Warning: not taken into accound WorkOrder transition/state.
        """
        if not self.entity.end_date or date.today() > self.entity.end_date:
            return date.today()
        return self.entity.end_date

    def completion_date(self):
        """returns date on which the subtask has been completed"""
        return date.today()

    def contractors(self):
        """returns the list of persons supposed to work on this task"""
        return []

    def in_progress(self):
        return self.entity.cw_adapt_to('IWorkflowable').state == u'in progress'

    def progress_info(self):
        return {'estimated': self.entity.budget,
                'estimatedcorrected': self.entity.budget,
                'done': self.entity.progress_done,
                'todo': self.entity.progress_todo}


# XXX IMileStone / IProgress inheritance stink
class WorkOrderProgressAdapter(WorkOrderIMileStoneAdapter):
    __regid__ = 'IProgress'
    __select__ = is_instance('WorkOrder')


class OrderProgressAdapter(IProgressAdapter):
    __select__ = is_instance('Order')

    def progress_info(self):
        return {'estimated': self.entity.budget,
                'estimatedcorrected': self.entity.budget,
                'done': self.entity.progress_done,
                'todo': self.entity.progress_todo,
                }

    def in_progress(self):
        return True  # FIXME


class OrderIMileStoneAdapter(IMileStoneAdapter):
    __select__ = is_instance('Order')

    def get_main_task(self):
        return self.entity

    def initial_prevision_date(self):
        """returns the initial expected end of the milestone"""
        return date.today()

    def eta_date(self):
        """returns expected date of completion based on what remains
        to be done
        """
        return date.today()

    def completion_date(self):
        """returns date on which the subtask has been completed"""
        return date.today()

    def contractors(self):
        """returns the list of persons supposed to work on this task"""
        return []

    def in_progress(self):
        return self.entity.cw_adapt_to('IWorkflowable').state == u'in progress'

    def progress_info(self):
        return {'estimated': self.entity.budget,
                'estimatedcorrected': self.entity.budget,
                'done': self.entity.progress_done,
                'todo': self.entity.progress_todo}



# postcreate script. You could setup a workflow here for example
from cubicweb import _

# XXX missing correct permissions configuration

# WorkOrder workflow
wwf = add_workflow(_('default workorder workflow'), 'WorkOrder')

reserve = wwf.add_state(_(u'not started'), initial=True)
encours = wwf.add_state(_(u'in progress'))
attente = wwf.add_state(_(u'client validation'))
garantie = wwf.add_state(_(u'warranty'))
recette = wwf.add_state(_(u'validated'))
annule = wwf.add_state(_(u'canceled'))

wwf.add_transition(_(u'cancel'), (reserve,), annule)
wwf.add_transition(_(u'start'), (reserve,), encours)
wwf.add_transition(_(u'done'), (encours,), attente)
wwf.add_transition(_(u'warrant'), (attente,), garantie)
wwf.add_transition(_(u'validated'), (attente, garantie), recette)


# Order workflow
owf = add_workflow(_('default order workflow'), 'Order')

planned = owf.add_state(_(u'planned'), initial=True)
canceled = owf.add_state(_(u'canceled'))
sent = owf.add_state(_(u'sent'))
in_progress = owf.add_state(_(u'in progress'))
done = owf.add_state(_(u'done'))
owf.add_transition(_(u'cancel'), (planned,sent), canceled)
owf.add_transition(_(u'send'), (planned,), sent)
owf.add_transition(_(u'receive'), (sent,), in_progress)
owf.add_transition(_(u'done'), (in_progress,), done)



from yams.buildobjs import SubjectRelation, String, Date, Float, RichString
from cubicweb.schema import WorkflowableEntityType

from cubicweb import _


class Order(WorkflowableEntityType):
    title = String(required=True, fulltextindexed=True, maxsize=128, unique=True)
    date = Date(description=_('date order was placed on'))
    split_into = SubjectRelation('WorkOrder', cardinality='*1', composite='subject')

    # computed attributes
    budget = Float(description=_('computed attribute'), required=True, default=0)
    progress_done = Float(description=_('computed attribute'))
    progress_todo = Float(description=_('computed attribute'))


class WorkOrder(WorkflowableEntityType):
    title = String(required=True, fulltextindexed=True, indexed=True, maxsize=128)
    description = RichString(fulltextindexed=True)
    budget = Float(description=_('amount that should not be overspent'))
    begin_date = Date(description=_('date work order will begin on'))
    end_date = Date(description=_('date work order will end on'))

    # computed attributes
    progress_done = Float(description=_('computed attribute'))
    progress_todo = Float(description=_('computed attribute'))



"""template-specific forms/views/actions/components"""

from cubicweb.predicates import is_instance
from cubicweb.web.views import uicfg
from cubicweb.web.views.ibreadcrumbs import IBreadCrumbsAdapter


class WorkOrderIBreadCrumbsAdapter(IBreadCrumbsAdapter):
    __select__ = is_instance('WorkOrder')

    def parent_entity(self):
        return self.entity.order


_pvs = uicfg.primaryview_section
_pvdc = uicfg.primaryview_display_ctrl
_afs = uicfg.autoform_section

# Order
_pvs.tag_subject_of(('Order', 'split_into', '*'), 'relations')
_pvs.tag_attribute(('Order', 'budget'), 'hidden')
_pvs.tag_attribute(('Order', 'progress_done'), 'hidden')
_pvs.tag_attribute(('Order', 'progress_todo'), 'hidden')
_pvdc.tag_subject_of(('Order', 'split_into', '*'),
                     {'vid': 'ic_progress_table_view'})

_afs.tag_attribute(('Order', 'budget'), 'main', 'hidden')
_afs.tag_attribute(('Order', 'progress_todo'), 'main', 'hidden')
_afs.tag_attribute(('Order', 'progress_done'), 'main', 'hidden')
_afs.tag_subject_of(('Order', 'split_into', '*'), 'main', 'inlined')

# WorkOrder
_afs.tag_attribute(('WorkOrder', 'progress_todo'), 'main', 'hidden')
_afs.tag_attribute(('WorkOrder', 'progress_done'), 'main', 'hidden')



# add attributes
add_attribute('Order','budget')
for attr in ('progress_target', 'progress_done', 'progress_todo'):
    add_attribute('Order', attr)
    add_attribute('WorkOrder', attr)
checkpoint()

from logilab.common.shellutils import ProgressBar
workorders = rql('Any X WHERE X is WorkOrder', ask_confirm=False)
bar = ProgressBar(len(workorders), title='Updating progress information ')
bar.refresh()
for entity in workorders.entities():
    entity.update_progress()
    bar.update()
print
checkpoint()

# add states
reserve = rql('Any S WHERE S is State, S name "not started", '
              'S state_of W, W name "WorkOrder"', ask_confirm=False)[0][0]
annule = add_state(_(u'canceled'), 'WorkOrder')
add_transition(_(u'cancel'), 'WorkOrder', (reserve,), annule)

planned = rql('Any S WHERE S is State, S name "planned", '
              'S state_of W, W name "Order"', ask_confirm=False)[0][0]
sent = rql('Any S WHERE S is State, S name "sent", '
           'S state_of W, W name "Order"', ask_confirm=False)[0][0]
canceled = add_state(_(u'canceled'), 'Order')
add_transition(_(u'cancel'), 'Order', (planned,sent), canceled)
checkpoint()
